
/*
  Copyright (C) 2010  Thomas Schrøder
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  LICENSE.txt file for license details.
*/

#include "rumd/rumd_algorithms.h"
#include "rumd/rumd_technical.h"
#include <sstream>

__host__ __device__ float4 operator+ (const float4 &a, const float4 &b){
  float4 r;
  r.x = a.x + b.x;
  r.y = a.y + b.y;
  r.z = a.z + b.z;
  r.w = a.w + b.w;
  return r;
}

__host__ __device__ void operator+= (float4 &a, const float4 &b){
  a.x += b.x;
  a.y += b.y;
  a.z += b.z;
  a.w += b.w;
}

__host__ __device__ void operator+= (volatile float4 &a, volatile const float4 &b){
  a.x += b.x;
  a.y += b.y;
  a.z += b.z;
  a.w += b.w;
}

__host__ __device__ double4 operator+ (const double4 &a, const double4 &b){
  double4 r;
  r.x = a.x + b.x;
  r.y = a.y + b.y;
  r.z = a.z + b.z;
  r.w = a.w + b.w;
  return r;
}

__host__ __device__ void operator+= (double4 &a, const double4 &b){
  a.x += b.x;
  a.y += b.y;
  a.z += b.z;
  a.w += b.w;
}

__host__ __device__ void operator+= (volatile double4 &a, volatile const double4 &b){
  a.x += b.x;
  a.y += b.y;
  a.z += b.z;
  a.w += b.w;
}


__host__ __device__ void operator+= (float2 &a, const float2 &b){
  a.x += b.x;
  a.y += b.y;
}

__host__ __device__ void operator+= (volatile float2 &a, volatile const float2 &b){
  a.x += b.x;
  a.y += b.y;
}

template <class T> std::string toString( T t, std::ios_base& (*f)(std::ios_base&) ){
  std::ostringstream oss;
  oss << f << t;
  return oss.str();
}

//////////////////////////////////////////////////////////
// Reduction algorithm: [Mark Harris, NVIDIA, SDK]
///////////////////////////////////////////////////////////

// Forward declaration.
template <class T, unsigned int blockSize> __global__ void reduce( T* g_idata, T* g_odata, unsigned int numParticles );

// Dynamically allocated shared memory is declared extern. Thus we need to define
// different names corresponding to different variable types.
template <typename T> struct SharedMemory{
  __device__ T* getPointer(){
    extern __device__ void error(void); // Undefined. Ensures that we won't compile any un-specialized types.
    error();
    return NULL;
  }
  __device__ T getZero(){ return NULL; }
};

template <> struct SharedMemory <float>{
  __device__ float* getPointer(){ extern __shared__ float s_float[]; return s_float; }    
  __device__ float getZero(){ return 0.f; }
};

template <> struct SharedMemory <float4>{
  __device__ float4* getPointer(){ extern __shared__ float4 s_float4[]; return s_float4; }    
  __device__ float4 getZero(){ float4 zero = { 0.f, 0.f, 0.f, 0.f }; return zero; }
};

template <> struct SharedMemory <double>{
  __device__ double* getPointer(){ extern __shared__ double s_double[]; return s_double; }    
  __device__ double getZero(){ return 0.; }
};

template <> struct SharedMemory <double4>{
  __device__ double4* getPointer(){ extern __shared__ double4 s_double4[]; return s_double4; }    
  __device__ double4 getZero(){ double4 zero = { 0., 0., 0., 0. }; return zero; }
};

template <> struct SharedMemory <float2>{
  __device__ float2* getPointer(){ extern __shared__ float2 s_float2[]; return s_float2; }    
  __device__ float2 getZero(){ float2 zero = { 0.f, 0.f }; return zero; }
};

inline void getThreadsAndBlocks( unsigned int size, unsigned int& numThreads, unsigned int& numBlocks, unsigned int maxThreadsPerBlock ){
  if(size == 1) 
    numThreads = 1;
  else
    numThreads = (size < 2 * maxThreadsPerBlock) ? size / 2 : maxThreadsPerBlock;
  
  numBlocks = size / (numThreads * 2);
}

template <class T> void sumIdenticalArrays( T* array, unsigned int numParticles, unsigned int numArrays, unsigned int maxThreadsPerBlock ){
  unsigned int numThreads = 0; unsigned int numBlocks = 0;
  unsigned int s = (unsigned) pow( 2, ceil( log( numParticles ) / log( 2 ) ) ); // Largest nearest power of 2.
  
  while(s > 1){
    getThreadsAndBlocks( s , numThreads, numBlocks, maxThreadsPerBlock );  
    dim3 dimGrid(numBlocks, numArrays);
    dim3 dimBlock(numThreads, 1);  
    
    // We must allocate 2 * numThreads * sizeof(T) due to warp unrolled in the reduction algorithm.
    unsigned int size = (numThreads <= 32) ? 2 * numThreads * sizeof(T) : numThreads * sizeof(T); 
    
    switch(numThreads){
    case 512:
      reduce<T,512><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 256:
      reduce<T,256><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 128:
      reduce<T,128><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 64:
      reduce<T,64><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 32:
      reduce<T,32><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 16:
      reduce<T,16><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 8:
      reduce<T,8><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 4:
      reduce<T,4><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 2:
      reduce<T,2><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    case 1:
      reduce<T,1><<< dimGrid, dimBlock, size >>>(array, array, numParticles); break;
    }
    
    s = s / (numThreads*2);
  }
}

// Performs a reduction.
template <class T, unsigned int blockSize> __global__ void reduce( T* g_idata, T* g_odata, unsigned int numParticles ){
  
  // Shared mem size is determined by the host app at run time.
  SharedMemory<T> smem;
  T* sdata = smem.getPointer();
  
  unsigned int tid = threadIdx.x;
  unsigned int i = blockIdx.x * (blockDim.x*2) + threadIdx.x;
  unsigned int offset = numParticles * blockIdx.y;
  
  T mySum = (i < numParticles) ? g_idata[offset+i] : smem.getZero();
  
  if( i + blockSize < numParticles ) 
    mySum += g_idata[offset+i+blockSize];  

  sdata[tid] = mySum;
  
  __syncthreads();
  
  // Unroll the initial iterations. We have to synchronize, since only 32 threads are atomic.
  if(blockSize >= 512) { if(tid < 256) { sdata[tid] += sdata[tid + 256]; } __syncthreads(); }
  if(blockSize >= 256) { if(tid < 128) { sdata[tid] += sdata[tid + 128]; } __syncthreads(); }
  if(blockSize >= 128) { if(tid <  64) { sdata[tid] += sdata[tid +  64]; } __syncthreads(); }
  
  if(tid < 32){
    volatile T* localSdata = sdata; // We must declare volatile amongst the 32 warp threads.
    
    // Unroll the last warp.
    if(blockSize >= 64) localSdata[tid] += localSdata[tid + 32]; 
    if(blockSize >= 32) localSdata[tid] += localSdata[tid + 16]; 
    if(blockSize >= 16) localSdata[tid] += localSdata[tid +  8]; 
    if(blockSize >= 8)  localSdata[tid] += localSdata[tid +  4]; 
    if(blockSize >= 4)  localSdata[tid] += localSdata[tid +  2]; 
    if(blockSize >= 2)  localSdata[tid] += localSdata[tid +  1]; 
  }
    
  // Write result for this block to global mem 
  if(tid == 0) 
    g_odata[offset+blockIdx.x] = sdata[0];
}

/////////////////////////////////////////////////////////////////////
// PRNG algorithm: [L. Howes and D. Thomas, GPU Gems 3, Chapter 37]
/////////////////////////////////////////////////////////////////////

__host__ __device__ inline unsigned TausStep( unsigned &z, int S1, int S2, int S3, unsigned M ){
  unsigned b = ( ( ( z << S1 ) ^ z ) >> S2 );
  return z = ( ( ( z & M ) << S3 ) ^ b );
}

__host__ __device__ inline unsigned LCGStep( unsigned &z, unsigned A, unsigned C ){
  return z = ( A * z + C );
}

// Different prefactor will produce (0, 1].
__host__ __device__ inline float HybridTaus( uint4* state ){
  return 2.3283064365387e-10f* (
				TausStep(state->x, 13, 19, 12, 4294967294UL) ^
				TausStep(state->y,  2, 25,  4, 4294967288UL) ^
				TausStep(state->z,  3, 11, 17, 4294967280UL) ^
				LCGStep( state->w,  1664525,   1013904223UL)
				);
}

/////////////////////////////////////////////////////////////////////
// Box Muller Transformation.
/////////////////////////////////////////////////////////////////////

__host__ __device__ inline float2 BoxMuller( float u0, float u1 ){
  float r = sqrtf( -2.0f * logf(u0) );
  float theta = 2.f * float(M_PI) * u1;
  float2 value = { r*sinf(theta), r*cosf(theta) };
  return value;
}

//////////////////////////////////////////////////////////////////////
// Gaussian Elimination with rowise (partial) pivoting and 
// backsubstitution
//////////////////////////////////////////////////////////////////////

template <class T> __global__ void GaussPivotBack( T* x, T* b, T* A, unsigned* dimlist ){
  unsigned i=0, j=0, k, kk, ir;
  unsigned lineIdx = (threadIdx.y*blockDim.x + threadIdx.x);
  unsigned lineDim = (blockDim.x*blockDim.y);
  unsigned nrc = dimlist[blockIdx.x]; // Number of constraints
  unsigned nra = nrc+1;
  T tmp;

  extern __shared__ T s_A[];
  T* s_c = &s_A[nra*nrc];

  // Load system from global to shared
  for ( k=lineIdx; k < nrc*nrc; k+=lineDim )
    s_A[ (k/nrc)*nra + k%nrc ] = A[blockIdx.x*nrc*nrc+k];
  for ( k=lineIdx; k < nrc; k+=lineDim )
    s_A[ k*nra + nrc ] = b[blockIdx.x*nrc+k];

  while ( i<nrc && j<nrc ){

    __syncthreads();

    // Find pivot
    ir = i;
    for ( k=i+1; k<nrc; k++ ){
      if ( fabs(s_A[k*nra +j]) > fabs(s_A[ir*nra+j]) )
	ir = k;
    }

    // Check that pivot is non-zero
    if ( fabs(s_A[ir*nra+j]) > FLT_EPSILON ){

      // Swap rows of s_A
      for ( k=lineIdx; k < nra; k+=lineDim ){
	tmp = s_A[i*nra+k];
	s_A[i*nra+k] = s_A[ir*nra+k];
	s_A[ir*nra+k] = tmp;
      }

      __syncthreads();

      // Temporary copy of pivot column
      for ( k=lineIdx; k < nrc; k+=lineDim )
	s_c[k] = s_A[k*nra+j];

      __syncthreads();

      // Divide row i with s_A(i,j)
      for ( k=lineIdx; k < nra; k+=lineDim )
	s_A[i*nra+k] /= s_c[i];

      // Multiply and subtract - Guassian part
      for ( k=threadIdx.x; k < nra; k+=blockDim.x )
	for ( kk=i+1+threadIdx.y; kk<nrc; kk+=blockDim.y )
	  s_A[kk*nra+k] -= s_c[kk]*s_A[i*nra+k];
   
      i++;
    }
    j++;
  }

  __syncthreads();

  // Backsubstitution
  for ( int i = nrc-1; i>=0; i-- )
    for ( k=lineIdx; k < i; k+=lineDim )
      s_A[k*nra+nrc] -= s_A[k*nra+i]*s_A[i*nra+nrc];

  __syncthreads();

  // Copy solution back to global
  for ( k=lineIdx; k < nrc; k+=lineDim )
    x[blockIdx.x*nrc+k] = s_A[k*nra+nrc];
}

template <class T> void solveLinearSystems( T* x, T* b, T* A, unsigned* dimLinearSystems, unsigned nLinearSystems, unsigned maxc){
  dim3 threadsPerBlock(maxc+1, 2);
  unsigned shared_size = maxc * (maxc + 2) * sizeof(float);

  GaussPivotBack<<<nLinearSystems, threadsPerBlock, shared_size>>>( x, b, A, dimLinearSystems );
}
